// ExamplePlugin.cpp : Defines the initialization routines for the DLL.
//

#include "DataSimulatorPlugin.h"

#if defined(OS_WIN32)
#include <windows.h>
#include <malloc.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "stat.h"

// constructor of your plugin class
//
DataSimulatorPlugin::DataSimulatorPlugin()
{
}

// destructor of your plugin class
//
DataSimulatorPlugin::~DataSimulatorPlugin()
{
}

bool DataSimulatorPlugin::initPlugin()
{
	// This function call and check of return value is necessary to check whether access to QX works
	//
	int ret = InitQXAccess();

	// Never continue, if the "InitQXAccess" function fails
	//
	if(!ret)
		return false;

	// Test whether the used QX version is new enough to support the commands used in your plugin
	//
	int Major, Minor, SubMinor;
	qxGetVersionOfBrainVoyagerQX(&Major, &Minor, &SubMinor);

	// We require v1.4.2
	if( !checkVersionGreaterEqualVersion(Major, Minor, SubMinor, 2, 0, 0) )
	{
		// we can use message box since this was available with the first QX version with plugin support (v1.3.8)
		qxMessageBox("Can not execute this plugin - BrainVoyagerQX version 2.0.0 or higher required.");
		return false;
	}

	return true;
}

// THE MOST IMPORTANT EXPORTED FUNCTION
//
bool DataSimulatorPlugin::execute()
{
	//
	// Now we can start with the plugin code
	//
	qxShowLogTab();
	
	qxLogText(" ");
	qxLogText("// Plugin>  Running Data Simulator...");

	// check whether we have a VMR, otherwise quit
	int DimXvmr, DimYvmr, DimZvmr;
	unsigned char *VMR = qxGetCurrentVMR(&DimXvmr, &DimYvmr, &DimZvmr);
	if(VMR == 0)
	{
		qxMessageBox("This plugin requires a VMR to operate. Make sure that a VMR is open and that its window is selected.");
		return false;
	}
	if( (DimXvmr != 256) || (DimYvmr != 256) || (DimZvmr != 256) )
	{
		qxMessageBox("This plugin requires a VMR with \"256\" dimensions, preferentially a VMR in Talairach space.");
		return false;
	}

	// Ask user how many time points he wants to simulate
	int NrOfSimulatedTimePoints = qxGetIntegerInput("Enter number of simulated time points:", 250, 1, 1000);
	if(NrOfSimulatedTimePoints < 1) // if value < min (1), user pressed cancel
		return false;
	qxCreateVTCForCurrentVMR(NrOfSimulatedTimePoints, 0); // since we provide no vtc header ("0" ptr), a float format VTC is created as default

	VTC_Header vtc_header;
	qxGetHeaderOfVTCOfCurrentVMR(&vtc_header);
	float *vtc = qxGetVTCFloatDataOfCurrentVMR();

	// determine dimension of VTC volume (Start and End values in header refer to position within VMR volume)
	int dim_x = (vtc_header.XEnd - vtc_header.XStart) / vtc_header.Resolution;
	int dim_y = (vtc_header.YEnd - vtc_header.YStart) / vtc_header.Resolution;
	int dim_z = (vtc_header.ZEnd - vtc_header.ZStart) / vtc_header.Resolution;
	int dim_xy  = dim_x * dim_y;
	int dim_xyz = dim_xy * dim_z;

	// check values by writing them to the Log
	char cText[260];
	sprintf( cText, "// Plugin>  Dims of created VTC, X: %i  Y: %i  Z: %i", dim_x, dim_y, dim_z );
	qxLogText( cText );

	//
	// Create cluster time courses based on protocol (or directly design matrix (RTC) file)
	//
	char PRTFile[261];
	int ret = qxGetOpenFileName("Provide a protocol file:", "", "Protocol Files (*.prt)", PRTFile );
	if( !ret )	return false; // user canceled
	strcpy(vtc_header.LinkedProtocolFile, PRTFile); // "CG_OBJECTS.prt");
	qxSetHeaderOfVTCOfCurrentVMR( &vtc_header );  // telling QX about updated VTC header results in linking of PRT to VTC
	// as an alternative, we can use qxLoadProtocol:
	//qxLoadProtocol(PRTFile);

	Protocol_Header prt_header;
	ProtocolCondition_Header prtcond_header;
	qxGetProtocol(&prt_header);
	int p, NrOfConditions = prt_header.NrOfConditions;


	// define predictors from all protocol conditions
	qxClearDesignMatrix();
	for(p=0; p<NrOfConditions; p++)
	{
		qxGetProtocolCondition(p, &prtcond_header);
		qxAddPredictor(prtcond_header.ConditionName); // predictor gets same name as condition from which it is created
		qxSetPredictorValuesFromCondition(prtcond_header.ConditionName, p, 1.0f);
		qxApplyTwoGammaHRFToPredictor(p);
	}

	// Alternatively, you may also prepare an RTC file in QX and load it directly (or ask user for a file name to load)
	//qxLoadSingleStudyGLMDesignMatrix("CG_OBJECTS.rtc");
	//float *sim_tc = qxGetPredictor(0);

	qxSaveSingleStudyGLMDesignMatrix("plugin_created.rtc");		// this fn is only called to demonstrate saving of created RTC

	//
	// Specify position/size etc. of cluster containing data
	//
	int PosX, PosY, PosZ, PosXvtc, PosYvtc, PosZvtc;
	int selected_condition, range = 2; // radius of spherical cluster, half-width of cubic cluster (defined already here since we restrict min/max values if user wants to enter coords)
	VOIs_Header vois_header;
	VOI_Header voi_header;
	qxGetVOIs(&vois_header);
	int v, NrOfVOIs = vois_header.NrOfVOIs;
	// variable to keep decision which time course to use for which VOI
	int *PredictorForVOI; PredictorForVOI = (int *)malloc(NrOfVOIs*sizeof(int)); if(PredictorForVOI == NULL) return false;
	bool UseVOIClusters = false;

	if(NrOfVOIs > 0)
	{
		UseVOIClusters = qxYesNoMessageBox("Do you want to use the available VOIs to define simulated clusters?");

		if(UseVOIClusters)
		{
			// ask for each VOI which predictor time course to use
			char ChooseFromConditionList[1001], ChooseInstruction[1100], ConditionNr[11];
			//sprintf(ChooseFromConditionList, "Select predictor:\n\n");
			strcpy(ChooseFromConditionList, "");
			for(p=0; p<NrOfConditions; p++)
			{
				qxGetProtocolCondition(p, &prtcond_header);
				sprintf(ConditionNr, "%i: ", (p+1));
				strcat(ChooseFromConditionList, ConditionNr);
				strcat(ChooseFromConditionList, prtcond_header.ConditionName);
				strcat(ChooseFromConditionList, "<br>");
			}

			for(v=0; v<NrOfVOIs; v++)
			{
				if( qxGetVOI(v, &voi_header) == 0) { free(PredictorForVOI); return false; }
				sprintf(ChooseInstruction, "Select predictor for VOI: <b>%s</b>:<br><br>", voi_header.VOIName);
				strcat(ChooseInstruction, ChooseFromConditionList);

				int selected_condition = qxGetIntegerInput(ChooseInstruction, 2, 1, NrOfConditions);
				if(selected_condition < 1)  { free(PredictorForVOI); return false; } // user cancelled dlg

				PredictorForVOI[v] = selected_condition - 1;  // adjust 1-based to 0-based value
			}
		}
	}

	if( !UseVOIClusters )  // if user has no VOIs defined/loaded, ask him to use cross position or ask him to enter coordinates
	{
		// which time course?
		//
		char ChooseFromConditionList[1001], ConditionNr[11];
		sprintf(ChooseFromConditionList, "Select condition:\n\n");
		for(p=0; p<NrOfConditions; p++)
		{
			qxGetProtocolCondition(p, &prtcond_header);
			sprintf(ConditionNr, "%i: ", (p+1));
			strcat(ChooseFromConditionList, ConditionNr);
			strcat(ChooseFromConditionList, prtcond_header.ConditionName);
			strcat(ChooseFromConditionList, "\n");
		}
		strcat(ChooseFromConditionList, "\n");

		selected_condition = qxGetIntegerInput(ChooseFromConditionList, 2, 1, NrOfConditions);
		if(selected_condition < 1) { free(PredictorForVOI); return false; } // user cancelled dlg

		selected_condition -= 1;  // adjust 1-based to 0-based value (var used below in data simulation loop)

		// Now get info about cluster posiiton:
		//
		ret = qxYesNoMessageBox("Do you want to use the cross position as the cluster center ?");
		if(ret)
			qxGetCrossPositionOfCurrentVMR(&PosX, &PosY, &PosZ);
		else
		{
			// let user specify center of cluster in VMR space (i.e. AC = 128, 128, 128)
			PosX = qxGetIntegerInput("Enter X position of simulated cluster: ", 128, vtc_header.XStart+range, vtc_header.XEnd-range);
			PosY = qxGetIntegerInput("Enter Y position of simulated cluster: ", 128, vtc_header.YStart+range, vtc_header.YEnd-range);
			PosZ = qxGetIntegerInput("Enter Z position of simulated cluster: ", 128, vtc_header.ZStart+range, vtc_header.ZEnd-range);
		}
		sprintf( cText, "// Plugin>  Selected position: X: %i  Y: %i  Z: %i", PosX, PosY, PosZ );
		qxLogText( cText );
	}

	int x, y, z, t, xyz;
	int n_vols = vtc_header.NrOfVolumes;

	float base_level = 100.0;
	float adj_psc = base_level / 100.0f;
	float max_psc = 3.0;		// max percent signal change of simulated signal (assuming time course between 0.0-1.0)
	float noise, noise_level = 5.0;
	float dist;
	long  idm = -1;

	// you can ask user whether he wants sphere or cube cluster shape
	bool  sphere_cluster = true;
	// you can ask user to use uniform noise or gaussian noise
	bool gauss_noise = true;

	sprintf( cText, "// Plugin>  Creating noise data...");
	qxLogText( cText );

	qxShowBusyCursor();

	for(t=0; t<n_vols; t++)
	{
		if( ((t%10) == 0) && (t>0) )
		{
			sprintf(cText, "// Plugin>    %i / %i ...", t, vtc_header.NrOfVolumes);
			qxLogText( cText );
		}

		for(z=0; z<dim_z; z++)
		{
			for(y=0; y<dim_y; y++)
			{
				for(x=0; x<dim_x; x++)
				{
					xyz = z*dim_xy + y*dim_x + x;

					if(gauss_noise)	noise = noise_level * gasdev(&idm);
					else			noise = noise_level * ran1(&idm) - noise_level/2.0f;

					// noise time courses
					vtc[xyz*n_vols + t] = base_level + noise;
				}
			}
		}
	}

	sprintf(cText, "// Plugin>    %i / %i ...", t, vtc_header.NrOfVolumes);
	qxLogText( cText );

	sprintf( cText, "// Plugin>  Creating signal + noise data...");
	qxLogText( cText );

	if( UseVOIClusters )
	{
		int vx, n_voxels, *voi_coords;

		for(v=0; v<NrOfVOIs; v++)
		{
			sprintf(cText, "// Plugin>    VOI %i / %i ...", v+1, NrOfVOIs);
			qxLogText( cText );

			// access the time course of the specified predictor for this VOI
			float *sim_tc = qxGetPredictor( PredictorForVOI[v] );
			if(sim_tc == NULL)
			{
				qxMessageBox("Error accessing predictor time course");
				free(PredictorForVOI);
				return false;
			}

			voi_coords = qxGetVOI(v, &voi_header);
			n_voxels = voi_header.NrOfVoxels;

			for(vx=0; vx<n_voxels; vx++)
			{
				if(vois_header.VOIsReferenceSpace == 3) // convert from TAL in BV system coordinates
				{
					PosX = 128 - voi_coords[3*vx+1]; // 128 -> ACxTAL);
					PosY = 128 - voi_coords[3*vx+2]; // 128 -> ACyTAL);
					PosZ = 128 - voi_coords[3*vx+0]; // 128 -> ACzTAL);
				}
				else
				{
					PosX = voi_coords[3*vx+0];
					PosY = voi_coords[3*vx+1];
					PosZ = voi_coords[3*vx+2];
				}

				// convert to position in VTC space
				int PosXvtc = qxRound( ((float)PosX-(float)vtc_header.XStart) / (float)vtc_header.Resolution );
				int PosYvtc = qxRound( ((float)PosY-(float)vtc_header.YStart) / (float)vtc_header.Resolution );
				int PosZvtc = qxRound( ((float)PosZ-(float)vtc_header.ZStart) / (float)vtc_header.Resolution );

				xyz = PosZvtc*dim_xy + PosYvtc*dim_x + PosXvtc;

				for(t=0; t<n_vols; t++)
				{
					if(gauss_noise)	noise = noise_level * gasdev(&idm);
					else			noise = noise_level * ran1(&idm) - noise_level/2.0f;

					// signal + noise time courses
					vtc[xyz*n_vols + t] = base_level + max_psc*adj_psc*sim_tc[t] + noise;
				}
			}
		}
	}
	else
	{
		// access the time course of the specified predictor for this VOI
		float *sim_tc = qxGetPredictor( selected_condition );
		if(sim_tc == NULL)
		{
			qxMessageBox("Error accessing predictor time course");
			free(PredictorForVOI);
			return false;
		}

		// convert to position in VTC space
		PosXvtc = qxRound( ((float)PosX-(float)vtc_header.XStart) / (float)vtc_header.Resolution );
		PosYvtc = qxRound( ((float)PosY-(float)vtc_header.YStart) / (float)vtc_header.Resolution );
		PosZvtc = qxRound( ((float)PosZ-(float)vtc_header.ZStart) / (float)vtc_header.Resolution );

		for(t=0; t<n_vols; t++)
		{
			for(z=PosZvtc-range; z<=PosZvtc+range; z++)
			{
				for(y=PosYvtc-range; y<=PosYvtc+range; y++)
				{
					for(x=PosXvtc-range; x<=PosXvtc+range; x++)
					{
						xyz = z*dim_xy + y*dim_x + x;

						if(sphere_cluster)
						{
							dist = sqrt( (float)((x-PosXvtc)*(x-PosXvtc) + (y-PosYvtc)*(y-PosYvtc) + (z-PosZvtc)*(z-PosZvtc)) );
							if(dist > (float)range)
								continue;
						}
			
						if(gauss_noise)	noise = noise_level * gasdev(&idm);
						else			noise = noise_level * ran1(&idm) - noise_level/2.0f;

						// signal + noise time courses
						vtc[xyz*n_vols + t] = base_level + (max_psc*adj_psc*sim_tc[t]) + noise;
					}
				}
			}
		}
	}

	

	qxStopBusyCursor();

	char VTCFileName[261];
	ret = qxGetSaveFileName("Enter Name For Saving Created VTC File:", "Simulated.vtc", "Volume Time Course Files (*.vtc)", VTCFileName );
	if( ret )
	{
		sprintf(cText, "// Plugin>  Saving created VTC file as \"%s\"...", VTCFileName);
		qxLogText( cText );
		qxShowBusyCursor();
		qxSaveVTCOfCurrentVMR( VTCFileName );
		qxStopBusyCursor();
	}

	qxLogText("// Plugin>  Data Simulator Plugin completed!");
	qxLogText(" ");

	qxUpdateActiveWindow();
	
	// free allocated heap memory
	free(PredictorForVOI);

	return true;
}

//
// C access functions (extern "C" in header)
//

// THIS FUNCTION CREATES YOUR PLUGIN CLASS
//
PLUGIN_ACCESS BVQXPluginInterface *createPlugin() 
{
	return new DataSimulatorPlugin();
}

// return a name for your plugin
//
PLUGIN_ACCESS const char *getPluginName() 
{
	return "Data Simulator";
}

// returns a descriptive string for your plugin shown in the QX "Descriptions of Plugins" window
//   it is recommended to use a static C string to describe your plugin
//   (the potentially lengthy string needs then not be created repeatedly in successive calls)
//   Note that you may use simple HTML tags to format the string, such as <b></b>, <i></i> and <br>
//
static const char *PluginDescription = "This tool allows to create simulated volume time course (VTC) data. Based on the position of the VMR cross or a set of defined Volumes-Of-Interest (VOIs), time courses are created with a time waveform based on selected conditions of a loaded protocol. The created voxel time courses are convolved with a hemodynamic response function. Finally Gaussian noise is added to all voxel time courses. The tool is implemented as a plugin and is provided with the source code to show how a fully-fledged plugin can be created. The simulated data should be used only to test data analysis methods for research and education.<br>";

PLUGIN_ACCESS const char *getPluginDescription() 
{
	return PluginDescription;
}

// get credit for your work!
//
PLUGIN_ACCESS const char *getAuthor()
{
	return "<i>Rainer Goebel</i>, Brain Innovation B.V.";
}

// you may provide the name for a HTML file containing online help
//
PLUGIN_ACCESS const char *getHelpFile()
{
	return "";
}

// you may provide a version string ("Major.Minor") for (updated) releases of your plugins
//
PLUGIN_ACCESS const char *getPluginVersion()
{
	return "2.0";
}

// NEW: If you want to use GUI dialogs, return a GUI script file, if not, return "" (or don't define this function)
//
PLUGIN_ACCESS const char *getGUIScript()
{
	return "";
}
